/*
 * File:    armour.c
 * Author:  Holger Klawitter
 * Date:    11.12.1998
 * Project: Cipher
 */

#include "armour.h"

Byte valid_bits = 0; /* number of valid bits in carry */
Word a = 0; /* accumulator */
Boolean higapped = false; /* true if armoured text does not use 0x80 - 0x9F */
Char* destPtr;

void packChar( Byte ch ) {
  a = a << 8 | ch;
  valid_bits += 8;
  while( valid_bits >= 7 ) {
    ch = '!' + ( a >> (valid_bits-7) & 0x7f );
		if( ch >= 0x80 ) ch += '!';
    *destPtr++ = ch;
    valid_bits -= 7;
    a &= (1 << valid_bits) - 1;
  }
}

void unpackChar( Byte ch ) {
  if( ch<'!' ) return;
	if( ch>=0x80 && higapped ) ch -= '!';
	if( ch>'!'+128) return; /* ignore non armour characters */
  a = ( a << 7 ) + ch - '!';
  valid_bits += 7;
  if( valid_bits >= 8 ) {
    *destPtr++ = ch = ( a >> (valid_bits-8) & 0xff );
    valid_bits -= 8;
    a &= (1<< valid_bits) - 1;
  }
}

/*
 * A byte sequence is being interpreted as a stream of bits.
 * The bits will be packed into blocks of 6 bits for each
 * character. Therefore the result string must be 8/6 times
 * as large as the source bit string.
 */
Char* armour( Byte* src, int n ) {
  Char* aText;
  int i;

  valid_bits = 0;
  a = 0;
  destPtr = aText = (Char*)MemPtrNew( sizeof(Char)*((n+6)*8/7) );

  for( i=0; i<n; i++ ) packChar( src[i] );

  /* There could be some bits still in the stream. Pushing one zero
   * into will surely emit them.
   */
  if( valid_bits ) packChar( 0 );
  *destPtr = 0;
  return aText;
}

/*
 * Reverses the process of packChars
 */
Byte* deArmour( Char* src, Int* len ) {
  Byte* sText;
  Int i;
  Int n = StrLen(src);
	Byte ch;

	// Determine armouring method
	higapped = true;
	for( i=0; i<n; i++ ) {
		ch = src[i];
		if( ch > 0x80 ) {
			higapped = ( ch >= 0xA0 );
			break;
		}
	}

  valid_bits = 0;
  a = 0;
  *len = ( ( (n*7/8)+7) / 8 ) * 8;  
  destPtr = sText = (Byte*)MemPtrNew( sizeof(Char)*(*len) );

  for( i=0; i<n; i++ ) unpackChar( src[i] );
  /* There might be odd bits in a, but the do not matter as packing already
   * has produced enough characters. If the user selected a sequence of wrong
   * size, she might therefore get a truncated result.
   */
  //*len = destPtr - sText;
  return sText;
}
