/*
 * File:    cipher.c
 * Author:  Holger Klawitter
 * Datum:   11.12.1998
 * Project: Cipher
 *
 * Changes:
 * 05.09.1999  HK  Password delay
 */
//#pragma pack(2)

#include <Common.h>
//#include <callback.h>
#include <System/SysAll.h>
#include <UI/UIAll.h>
#include "cipher.h"
#include "cipher.rch"
#include "crypt.h"

/***********************************************************************
 * Helper functions
 ***********************************************************************/

VoidPtr GetObjectPtr( Word objectId ) {
  FormPtr frm;
  frm = FrmGetActiveForm();
  return FrmGetObjectPtr( frm, FrmGetObjectIndex( frm, objectId ) );
}

Word GetObjectIdx( Word objectId ) {   
  FormPtr frm;
  frm = FrmGetActiveForm();
  return FrmGetObjectIndex( frm, objectId );
}

/***********************************************************************
 * Event Handler for the About Form
 ***********************************************************************/
static Boolean AboutFormHandleEvent( EventPtr eptr )
{
  Boolean handled = false;

  switch( eptr->eType ) {
  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {
    case AboutOkButton:
      FrmGotoForm( MainForm );
      handled = true;
      break;
    }
    break;
  case frmOpenEvent:
    FrmDrawForm(FrmGetActiveForm());
    handled = true;
    break;

  default:
  }

  return handled;
}


/***********************************************************************
 * Actual code
 ***********************************************************************/

Char* theKey = NULL;

/*
 * Compute Key to use
 */
void ComputeKey() {
  FieldPtr fld = GetObjectPtr( KeyField );
  VoidHand handle = (VoidHand)FldGetTextHandle( fld );
  CharPtr text;
  Int len;

  if( !FldDirty( fld ) && CtlGetValue( GetObjectPtr(KeyCheck) ) ) {
    return; /* keeps old key if checked and no new key present */
  }
  if( !handle ) return; /* no key */
  if( theKey ) {
    MemSet( theKey, StrLen(theKey), 0 );
    MemPtrFree( theKey );
  }
  text = MemHandleLock( handle );
  len = StrLen(text);
  theKey = (Char*)MemPtrNew( sizeof(Char)*(len+1) );
  StrCopy( theKey, text );
  MemHandleUnlock( handle );
}

/*
 * Removes the key from display, and selects the key check
 * to indicate the presence of the key.
 */
void CleanKey() {
  FieldPtr fld = GetObjectPtr( KeyField );
  CharPtr str = FldGetTextPtr(fld);
  if( str ) {
    FldDelete( fld, 0, StrLen(str) );
    FldSetDirty( fld, false );
  }
  CtlSetValue( GetObjectPtr(KeyCheck), true );
}

#ifndef DebugMsg
Int DebugMsg( CharPtr msg ) {
  FrmCustomAlert( Debug, msg, NULL, NULL );
  return 0;
}
#endif

/*
 * Replaces the text in the field with the new text.
 */
void ReplaceText( FieldPtr fld, CharPtr text ) {
  VoidHand oldHandle = (VoidHand)FldGetTextHandle( fld );
  VoidHand newHandle = (VoidHand)MemHandleNew( sizeof(Char)*4096 );
  CharPtr newHText = MemHandleLock( newHandle );
  CharPtr oldText;

  if( text ) {
    MemMove( newHText, text, StrLen( text )+1 );
  } else {
    newHText[0] = 0;
  }
  MemHandleUnlock( newHandle );
  
  oldText = MemHandleLock( oldHandle );
  MemSet( oldText, StrLen(oldText), 0 );
  MemHandleUnlock( oldHandle );

  FldSetTextHandle( fld, (Handle)newHandle );
  MemHandleFree( oldHandle );
  FldDrawField( fld );
}

/*
 * Takes the contents of the Textfield and enciphers it.
 */
void Cipher( Boolean weak ) {
  FieldPtr fld = GetObjectPtr( TextField );
  VoidHand handle = (VoidHand)FldGetTextHandle( fld );
  CharPtr text,cText;

  if( !handle ) return; /* nothing to do */
  ComputeKey();
  text = MemHandleLock( handle );
  if( StrLen(text)>=MAX_TEXT_LEN ) {
    FrmAlert( TooLongError );
    MemHandleUnlock( handle );
    return;
  }
  cText = encrypt( theKey, text, weak );
  MemHandleUnlock( handle );
  if( cText ) ReplaceText( fld, cText );
  CleanKey();
  CtlSetValue( GetObjectPtr(CopyCheck), true );
}

/*
 * Takes the contents of the Textfield and deciphers it.
 */
void Decipher( Boolean weak ) {
  FieldPtr fld = GetObjectPtr( TextField );
  VoidHand handle = (VoidHand)FldGetTextHandle( fld );
  CharPtr text, oText;

  if( !handle ) return; /* nothing to do */
  ComputeKey();
  text = MemHandleLock( handle );
  oText = decrypt( theKey, text, weak );
  MemHandleUnlock( handle );
  if( !oText ) {
    FrmAlert( BadKeyError );
    return;
  }
  ReplaceText( fld, oText );
  CleanKey();
  CtlSetValue( GetObjectPtr(CopyCheck), false );
}

void CopyFieldToBuffer() {
  FieldPtr fld = GetObjectPtr( TextField );
  VoidHand handle = (VoidHand)FldGetTextHandle( fld );
  CharPtr text;
  Int len, mark;

  if( !handle ) return; /* nothing to do */
  text = MemHandleLock( handle );
  len = StrLen(text);
  if( CtlGetValue(GetObjectPtr(CopyCheck)) ) {
    mark = len;
  } else {
    MemSet( text, len, 0 ); // clean data
    mark = 0;
  }
  MemHandleUnlock( handle );
  FldSetSelection( fld, 0, mark );
  FldCopy( fld );
}

/***********************************************************************
 * Event Handler for the Main Form
 ***********************************************************************/

static void DoScroll( DirectionType dir, Word amount ) {
  FieldPtr fld = GetObjectPtr( TextField );
  if( FldScrollable( fld, dir ) ) {
    FldScrollField( fld, amount, dir );
  }
}

static void FieldCommand( UInt cmd, FieldPtr fld ) {
	if( !fld ) { return; }
	switch( cmd ) {
	case EditUndo:
		FldUndo( fld );
		break;
	case EditCut:
		FldCut( fld );
		break;
	case EditCopy:
		FldCopy( fld );
		break;
	case EditPaste:
		FldPaste( fld );
		break;
	case EditSelectAll:
		FldSetSelection( fld, 0, FldGetTextLength(fld) );
		break;
	case EditKeyboard:
		SysKeyboardDialog();
		break;
	}
}

static Boolean MainFormHandleEvent( EventPtr eptr ) {
  Boolean  handled = false;
  FormPtr  frm = FrmGetActiveForm();
	Int      idx;

  //CALLBACK_PROLOGUE;
  switch( eptr->eType ) {

  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {

    case CipherButton:
      Cipher( false );
      handled = true;
      break;
      
    case DecipherButton:
      Decipher( false );
      handled = true;
      break;

    case UpButton:
      DoScroll( up, 1 );
      handled=true;
      break;

    case DownButton:
      DoScroll( down, 1 );
      handled=true;
      break;

    }
    break;

  case keyDownEvent:
    switch( eptr->data.keyDown.chr ) {

    case pageUpChr:
      DoScroll( up, 9 );
      handled = true;
      break;

    case pageDownChr:
      DoScroll( down, 9 );
      handled = true;
      break;

    }
    break;

  case menuEvent:
    switch( eptr->data.menu.itemID ) {
			
		case EditUndo:
		case EditCut:
		case EditCopy:
		case EditPaste:
		case EditSelectAll:
		case EditKeyboard:
			idx = FrmGetFocus( frm );
			if( idx != -1 ) {
				FieldCommand( eptr->data.menu.itemID,	FrmGetObjectPtr( frm, idx ) );
			}
			handled = true;
			break;

		case CipherEncrypt:
			Cipher( false );
			handled = 1;
			break;

		case CipherDecrypt:
			Decipher( false );
			handled = 1;
			break;

		case CipherWEncrypt:
			Cipher( true );
			handled = 1;
			break;

		case CipherWDecrypt:
			Decipher( true );
			handled = 1;
			break;

    case CipherAbout:
      FrmPopupForm( AboutForm );
      handled = true;
      break;
    }
    break;

  case frmOpenEvent:
    FrmDrawForm( frm );
    FldPaste( GetObjectPtr( TextField ) );
    FrmSetFocus( frm, FrmGetObjectIndex( frm, KeyField ) );
    handled = true;
    break;

  case appStopEvent:
    DebugMsg( "stopping" );
    CopyFieldToBuffer();
    DebugMsg( "stopped" );
    break;

  default:
  }
  //CALLBACK_EPILOGUE;

  return handled;
}

/***********************************************************************
 * MAIN PROGRAM STUFF
 ***********************************************************************/

typedef struct {
	Char key[16];
} Prefs;

/***********************************************************************
 * Starting and ending the application
 **********************************************************************/
static Word StartApplication()
{
  return 0;
}

static void StopApplication()
{
  if( theKey ) MemSet( theKey, StrLen(theKey), 0 );
  if( ks ) MemSet( ks, sizeof(Word)*INTERNAL_KEY_LENGTH, 0 );
}

/***********************************************************************
 * The master Event handler stuff
 ***********************************************************************/
static Boolean ApplicationHandleEvent( EventPtr eptr )
{
  FormPtr frm;
  Int formId;
  Boolean handled = false;

  if( eptr->eType == frmLoadEvent ) {
    formId = eptr->data.frmLoad.formID;
    frm = FrmInitForm( formId );
    FrmSetActiveForm( frm );

    switch( formId ) {
    case MainForm:
      FrmSetEventHandler( frm, (MainFormHandleEvent) );
      break;
    case AboutForm:
      FrmSetEventHandler( frm, (AboutFormHandleEvent) );
      break;
    }
    handled = true;
  }  
  return handled;
}

static void EventLoop()
{
  EventType e;
  Word err;

  do {
    EvtGetEvent( &e, evtWaitForever );
    if( !SysHandleEvent( &e ) &&
	!MenuHandleEvent( NULL, &e, &err ) &&
	!ApplicationHandleEvent( &e )
    ) {
      FrmDispatchEvent( &e );
    }
  } while( e.eType != appStopEvent );
}

/***********************************************************************
 * The Main loop
 ***********************************************************************/
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Word err;

  if( cmd == sysAppLaunchCmdNormalLaunch ) {
    err = StartApplication();
    if( err ) {
      return err;
    }
    FrmGotoForm( MainForm );
    EventLoop();
    StopApplication();
  }

  return 0;
}
