/*
 * File:    crypt.c
 * Author:  Holger Klawitter
 * Date:    11.12.1998
 * Project: Cipher 
 */

#include "armour.h"
#include "crypt.h"
#include "cipher.h"

//////////////////////////////////////////////////////////////////////
// The IDEA(tm) algorithm is copyrighted and trademarked by Ascom
// The folloing implementation is copyrighted by the
// Cryptics Development Team (Systemics, Ltd.)
//////////////////////////////////////////////////////////////////////

Word* ks = 0;

Word mul (ULong a, ULong b) {
  ULong p;
  Word res;

  a &= 0xFFFF;
  b &= 0xFFFF;
  if (a != 0) {
    if (b != 0) {
      p = a * b;
      b = p & 0xFFFF;
      a = p >> 16;
      res = (Word)(b - a + (b < a ? 1 : 0));
    } else
      res = (Word)(1 - a);
  } else
    res = (Word)(1 - b);
  //printf( "%04X * %04X = %04X\n", a, b, res );
  return res;
}

Word inv (Long xx) {
  Long x,t1,y,t0,q;

  x = xx & 0xFFFF;     // only lower 16 bits
  if (x <= 1)
    return (Word)x;       // 0 and 1 are self-inverse
  
  t1 = 0x10001L / x;    // Since x >= 2, this fits into 16 bits
  y = 0x10001L % x;
  if (y == 1)
    return (Word)(1 - t1);
  
  t0 = 1;
  do {
    q = x / y;
    x = x % y;
    t0 += q * t1;
    if (x == 1)
      return (Word)t0;
    q = y / x;
    y %= x;
    t1 += q * t0;
  } while (y != 1);
  return (Word)(1 - t1);
}

void loadKey( Byte* userkey ) {
  Int i;
  Int zoff;
  Int j;

  if( !ks ) {
    ks = (Word*)MemPtrNew( sizeof(Word)*INTERNAL_KEY_LENGTH );
  }

  ks[0] = ((Word)(userkey[ 0] & 0xFF) << 8 | (userkey[ 1] & 0xFF));
  ks[1] = ((Word)(userkey[ 2] & 0xFF) << 8 | (userkey[ 3] & 0xFF));
  ks[2] = ((Word)(userkey[ 4] & 0xFF) << 8 | (userkey[ 5] & 0xFF));
  ks[3] = ((Word)(userkey[ 6] & 0xFF) << 8 | (userkey[ 7] & 0xFF));
  ks[4] = ((Word)(userkey[ 8] & 0xFF) << 8 | (userkey[ 9] & 0xFF));
  ks[5] = ((Word)(userkey[10] & 0xFF) << 8 | (userkey[11] & 0xFF));
  ks[6] = ((Word)(userkey[12] & 0xFF) << 8 | (userkey[13] & 0xFF));
  ks[7] = ((Word)(userkey[14] & 0xFF) << 8 | (userkey[15] & 0xFF));
  
  for ( i = 0, zoff = 0, j = 8; j < INTERNAL_KEY_LENGTH; i &= 7, j++) {
    i++;
    ks[i + 7 + zoff] =
      (Word)(((ULong)(ks[(i & 7) + zoff]) << 9) |
	      (((ULong)(ks[((i + 1) & 7) + zoff]) >> 7) & 0x1FF));
    zoff += i & 8;
  }
}

void invertKey () {
  int i, j = 4, k = INTERNAL_KEY_LENGTH - 1;
  Word temp[INTERNAL_KEY_LENGTH];

  temp[k--] = inv(ks[3]);
  temp[k--] = (Word) -ks[2];
  temp[k--] = (Word) -ks[1];
  temp[k--] = inv(ks[0]);
  for (i = 1; i < ROUNDS; i++, j += 6) {
    temp[k--] = ks[j + 1];
    temp[k--] = ks[j];
    temp[k--] = inv(ks[j + 5]);
    temp[k--] = (Word) -ks[j + 3];
    temp[k--] = (Word) -ks[j + 4];
    temp[k--] = inv(ks[j + 2]);
  }
  temp[k--] = ks[j + 1];
  temp[k--] = ks[j];
  temp[k--] = inv(ks[j + 5]);
  temp[k--] = (Word) -ks[j + 4];
  temp[k--] = (Word) -ks[j + 3];
  temp[k--] = inv(ks[j + 2]);
  MemMove( ks, temp, sizeof(Word)*INTERNAL_KEY_LENGTH );
}

void blockEncrypt (Byte* text) {
  Word
    x1 = (Word)(((text[0] & 0xFF) << 8) | (text[1] & 0xFF)),
    x2 = (Word)(((text[2] & 0xFF) << 8) | (text[3] & 0xFF)),
    x3 = (Word)(((text[4] & 0xFF) << 8) | (text[5] & 0xFF)),
    x4 = (Word)(((text[6] & 0xFF) << 8) | (text[7] & 0xFF)),
    s2, s3;
  int i = 0,
    round = ROUNDS;
  
  while (round-- > 0) {
    x1 = mul(x1, ks[i++]);
    x2 += ks[i++];
    x3 += ks[i++];
    x4 = mul(x4, ks[i++]);
    
    s3 = x3;
    x3 = mul(x1 ^ x3, ks[i++]);
    s2 = x2;
    x2 = mul(x3 + (x2 ^ x4), ks[i++]);
    x3 += x2;
    
    x1 ^= x2;
    x4 ^= x3;
    x2 ^= s3;
    x3 ^= s2;
  }
  s2 = mul(x1, ks[i++]);
  text[0] = (Byte)(s2 >> 8);
  text[1] = (Byte) s2;
  s2 = (Word)(x3 + ks[i++]);
  text[2] = (Byte)(s2 >> 8);
  text[3] = (Byte) s2;
  s2 = (Word)(x2 + ks[i++]);
  text[4] = (Byte)(s2 >> 8);
  text[5] = (Byte) s2;
  s2 = mul(x4, ks[i]);
  text[6] = (Byte)(s2 >> 8);
  text[7] = (Byte) s2;
}

//////////////////////////////////////////////////////////////////////
// Foreign code ends here.
//////////////////////////////////////////////////////////////////////

void preparePrimaryKey( Char* key ) {
  Char longerKey[KEY_LENGTH];
  
  if( key && StrLen(key) >= KEY_LENGTH ) {
    loadKey( key );
  } else {
    MemSet( longerKey, KEY_LENGTH, 0 );
    if( key ) StrCopy( longerKey, key );
    loadKey( longerKey );
    MemSet( longerKey, KEY_LENGTH, 0 ); // dont leave clean text behind!
  }
}

void prepareKey( Char* key, Boolean weakEncrypt ) {
	CharPtr realKey;
	UInt    keyLen;
	UInt    realKeyLen;
	UInt    padding;
	Int     i,j;

	// create a weaker key to initiate encrypted "real" key.
	preparePrimaryKey( key );
	if( weakEncrypt ) { return; }

	// create clean text key with matching blocklen
	// and do some nice padding
	keyLen = StrLen(key);
	realKeyLen = ( keyLen + KEY_LENGTH-1 ) & 0xfff0;

	padding = realKeyLen - keyLen;
	realKey = MemPtrNew( sizeof(Char)*realKeyLen );
	MemSet( realKey, realKeyLen, padding + (padding << 4) );
	StrCopy( realKey, key );
	
	// encrypt key with weak key and cipher block chaining
	// BLOCK_SIZE must divide KEY_LENGTH
	for( i=0; i<=realKeyLen-BLOCK_SIZE; i+=BLOCK_SIZE ) {
		if( i>0 ) {
			for( j=0; j<BLOCK_SIZE; j++ ) {
				realKey[i+j] ^= realKey[i+j-BLOCK_SIZE];
			}
		}
		blockEncrypt( realKey + i );
	}

	// get the last part as the real key
	i -= KEY_LENGTH;
	loadKey( realKey+i );

	// cleanup
	MemSet( realKey, realKeyLen, 0 );
	MemPtrFree( realKey );
}

static void StrNCopy( CharPtr dest, CharPtr src, Word n ) {
	while( n-->0 && *src ) {
		*dest++ = *src++;
	}
}

Char* encrypt( Char* key, Char* text, Boolean weakEncrypt ) {
  Byte lastBlock[BLOCK_SIZE];
  Byte* sText;
  Char* aText;
  Int i,j;
  Int len, sLen;

  if( !text || !key || !(len=StrLen(text)) || !StrLen(key) ) return NULL;
  sLen = ((len+10)/8)*8; // + 3 bytes, round up to multiple of 8

	// bring text to blocklength and encode length of string into it
  sText = (Byte*)MemPtrNew( sizeof(Byte)*sLen );
  MemSet( sText, sLen, 0 );
  StrNCopy( sText+3, text, len );
  sText[0] = len & 0xFF;
  sText[1] = len >> 8;
  sText[2] = SysRandom(0);

  prepareKey( key, weakEncrypt );

  // for( i=0; i<sLen; i++ ) { printf( "(%d)", sText[i] ); } printf( "\n" );
  MemSet( lastBlock, BLOCK_SIZE, 0 );
  for( i=0; i<sLen; i+=BLOCK_SIZE ) {
    for( j=0; j<BLOCK_SIZE; j++ )
      lastBlock[j] ^= sText[i+j];
    blockEncrypt( lastBlock );
    MemMove( sText+i, (Char*)lastBlock, BLOCK_SIZE );
  }
  //for( i=0; i<sLen; i++ ) { printf( "(%d)", sText[i] ); } printf( "\n" );

  aText = armour( sText, sLen );
  MemSet( sText, sLen, 0 );
  MemPtrFree( sText );
  return aText;
}

Char* decrypt( Char* key, Char* aText, Boolean weakEncrypt ) {
  Char lastBlock[BLOCK_SIZE];
  Char curBlock[BLOCK_SIZE];
  Byte* sText;
  Char* text;
  Int i,j;
  Int len,sLen;

  if( !aText || !key || StrLen(aText)<2 || !StrLen(key) ) return NULL;

	// remove armouring, now the text may contain zeros!
  sText = deArmour( aText, &sLen );

  // load key into idea
  prepareKey( key, weakEncrypt );
  invertKey();

  //for( i=0; i<sLen; i++ ) { printf( "(%d)", sText[i] ); } printf( "\n" );
  MemSet( lastBlock, BLOCK_SIZE, 0 );
  for( i=0; i<sLen; i+=BLOCK_SIZE ) {
    MemMove( curBlock, sText+i, BLOCK_SIZE );
    blockEncrypt( sText+i );
    for( j=0; j<BLOCK_SIZE; j++ )
      sText[i+j] ^= lastBlock[j];
    MemMove( lastBlock, curBlock, BLOCK_SIZE );
  }
  //for( i=0; i<sLen; i++ ) { printf( "(%d)", sText[i] ); } printf( "\n" );

  // readout text in correct length
  len = (Byte)(sText[1]) << 8 | (Byte)sText[0];
  if( len>0 && len+3 <= sLen ) {
    text = (Char*)MemPtrNew( sizeof(Char)*(len+1) );
    MemMove( text, sText+3, len );
    text[len] = 0;
  } else {
    //printf( "apparently bad key!\n" );
    text = NULL;
  }

  // cleanup
  MemSet( sText, sLen, 0 );
  MemPtrFree( sText );

  return text;
}
